import dataclasses
import struct
from enum import Enum
from pathlib import Path
from typing import BinaryIO, Optional, Tuple, Iterable

from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.games.game import RandovaniaGame
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences

_NUM_TEXT_SECTIONS = 7
_NUM_DATA_SECTIONS = 11
_NUM_SECTIONS = _NUM_TEXT_SECTIONS + _NUM_DATA_SECTIONS


@dataclasses.dataclass(frozen=True)
class Section:
    offset: int
    base_address: int
    size: int


@dataclasses.dataclass(frozen=True)
class DolHeader:
    sections: Tuple[Section, ...]
    bss_address: int
    bss_size: int
    entry_point: int

    @classmethod
    def from_bytes(cls, data: bytes) -> "DolHeader":
        struct_format = f">{_NUM_SECTIONS}L"
        offset_for_section = struct.unpack_from(struct_format, data, 0)
        base_address_for_section = struct.unpack_from(struct_format, data, 0x48)
        size_for_section = struct.unpack_from(struct_format, data, 0x90)

        bss_address, bss_size, entry_point = struct.unpack_from(">LLL", data, 0xD8)
        sections = tuple(Section(offset_for_section[i], base_address_for_section[i], size_for_section[i])
                         for i in range(_NUM_SECTIONS))

        return cls(sections, bss_address, bss_size, entry_point)

    def offset_for_address(self, address: int) -> Optional[int]:
        for section in self.sections:
            relative_to_base = address - section.base_address
            if 0 <= relative_to_base < section.size:
                return section.offset + relative_to_base
        return None


class DolFile:
    dol_file: Optional[BinaryIO] = None
    editable: bool = False

    def __init__(self, dol_path: Path):
        with dol_path.open("rb") as f:
            header_bytes = f.read(0x100)

        self.dol_path = dol_path
        self.header = DolHeader.from_bytes(header_bytes)

    def set_editable(self, editable: bool):
        self.editable = editable

    def __enter__(self):
        if self.editable:
            f = self.dol_path.open("r+b")
        else:
            f = self.dol_path.open("rb")
        self.dol_file = f.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dol_file.__exit__(exc_type, exc_type, exc_tb)
        self.dol_file = None

    def offset_for_address(self, address: int) -> int:
        offset = self.header.offset_for_address(address)
        if offset is None:
            raise ValueError(f"Address 0x{address:x} could not be resolved for dol at {self.dol_file}")
        return offset

    def read(self, address: int, size: int) -> bytes:
        offset = self.offset_for_address(address)
        self.dol_file.seek(offset)
        return self.dol_file.read(size)

    def write(self, address: int, code_points: Iterable[int]):
        offset = self.offset_for_address(address)
        self.dol_file.seek(offset)
        self.dol_file.write(bytes(code_points))


@dataclasses.dataclass(frozen=True)
class StringDisplayPatchAddresses:
    update_hint_state: int
    message_receiver_string_ref: int
    wstring_constructor: int
    display_hud_memo: int
    cstate_manager_global: int
    max_message_size: int


@dataclasses.dataclass(frozen=True)
class HealthCapacityAddresses:
    base_health_capacity: int
    energy_tank_capacity: int


@dataclasses.dataclass(frozen=True)
class BeamCostAddresses:
    uncharged_cost: int
    charged_cost: int
    charge_combo_ammo_cost: int
    charge_combo_missile_cost: int
    get_beam_ammo_type_and_costs: int


@dataclasses.dataclass(frozen=True)
class SafeZoneAddresses:
    heal_per_frame_constant: int
    increment_health_fmr: int


@dataclasses.dataclass(frozen=True)
class PatchesForVersion:
    game: RandovaniaGame
    description: str
    build_string_address: int
    build_string: bytes
    sda2_base: int
    game_state_pointer: int
    string_display: StringDisplayPatchAddresses
    health_capacity: HealthCapacityAddresses
    beam_cost_addresses: BeamCostAddresses
    game_options_constructor_address: int
    safe_zone: SafeZoneAddresses


ALL_VERSIONS_PATCHES = [
    PatchesForVersion(
        game=RandovaniaGame.PRIME2,
        description="Gamecube NTSC",
        build_string_address=0x803ac3b0,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32",
        sda2_base=0x804223c0,
        game_state_pointer=0x80418eb8,
        string_display=StringDisplayPatchAddresses(
            update_hint_state=0x80038020,
            message_receiver_string_ref=0x803bd118,
            wstring_constructor=0x802ff3dc,
            display_hud_memo=0x8006b3c8,
            cstate_manager_global=0x803db6e0,
            max_message_size=200,
        ),
        health_capacity=HealthCapacityAddresses(
            base_health_capacity=0x8041abe4,
            energy_tank_capacity=0x8041abe0,
        ),
        beam_cost_addresses=BeamCostAddresses(
            uncharged_cost=0x803aa8c8,
            charged_cost=0x803aa8d8,
            charge_combo_ammo_cost=0x803aa8e8,
            charge_combo_missile_cost=0x803a74ac,
            get_beam_ammo_type_and_costs=0x801cccb0,
        ),
        game_options_constructor_address=0x80161b48,
        safe_zone=SafeZoneAddresses(
            heal_per_frame_constant=0x8041a4fc,
            increment_health_fmr=0x8000c710,
        ),
    ),
    PatchesForVersion(
        game=RandovaniaGame.PRIME2,
        description="Gamecube PAL",
        build_string_address=0x803ad710,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.035 10/27/2004 19:48:17",
        sda2_base=0x804223c0,
        game_state_pointer=0x8041A19C,
        string_display=StringDisplayPatchAddresses(
            update_hint_state=0x80038194,
            message_receiver_string_ref=0x803be378,
            wstring_constructor=0x802ff734,
            display_hud_memo=0x8006b504,
            cstate_manager_global=0x803dc900,
            max_message_size=200,
        ),
        health_capacity=HealthCapacityAddresses(
            base_health_capacity=0x8041bedc,
            energy_tank_capacity=0x8041bed8,
        ),
        beam_cost_addresses=BeamCostAddresses(
            uncharged_cost=0x803abc28,
            charged_cost=0x803abc38,
            charge_combo_ammo_cost=0x803abc48,
            charge_combo_missile_cost=0x803a7c04,
            get_beam_ammo_type_and_costs=0x801ccfe4,
        ),
        game_options_constructor_address=0x80161d9c,
        safe_zone=SafeZoneAddresses(
            heal_per_frame_constant=0x8041b7f4,
            increment_health_fmr=0x8000c754,
        ),
    ),
]

_PREFERENCES_ORDER = (
    "sound_mode",
    "screen_brightness",
    "screen_x_offset",
    "screen_y_offset",
    "screen_stretch",
    "sfx_volume",
    "music_volume",
    "hud_alpha",
    "helmet_alpha",
)
_FLAGS_ORDER = (
    "hud_lag",
    "invert_y_axis",
    "rumble",
    None,  # crashes, maybe swapBeamsControls
    "hint_system",
    None,  # 5: doesn't crash, unknown
    None,  # 6: doesn't crash, unknown
    None,  # 7: doesn't crash, unknown
)


def _apply_string_display_patch(patch_addresses: StringDisplayPatchAddresses, dol_file: DolFile):
    message_receiver_string_ref = struct.pack(">I", patch_addresses.message_receiver_string_ref)
    address_wstring_constructor = struct.pack(">I", patch_addresses.wstring_constructor)
    address_display_hud_memo = struct.pack(">I", patch_addresses.display_hud_memo)

    # setup stack
    # stwu r1,-0x2C(r1)
    # mfspr r0,LR
    # stw r0,0x30(r1)

    # # return if displayed
    # lbz r4,0x2(r3)
    # cmpwi r4,0x0
    # beq end

    # # otherwise set displayed
    # li r6,0x0
    # stb r6,0x2(r3)

    # # setup CHUDMemoParms
    # lis r5,0x4100 # 8.0f
    # li r7,0x1
    # li r9,0x9
    # stw r5,0x10(r1) # display time (seconds)
    # stb r7,0x14(r1) # clear memo window
    # stb r6,0x15(r1) # fade out only
    # stb r6,0x16(r1) # hint memo
    # stb r7,0x17(r1) # fade in text
    # stw r9,0x18(r1) # unk

    # # setup wstring
    # addi r3,r1,0x1C
    # lis r4,0x803a      # string pointer
    # ori r4,r4,0x6380
    # lis r12,0x802f     # wstring_l constructor
    # ori r12,r12,0xf3dc
    # mtspr CTR,r12
    # bctrl # rstl::wstring_l

    # # r3 = wstring
    # addi r4,r1,0x10
    # lis r12,0x8006     # DisplayHudMemo address
    # ori r12,r12,0xb3c8
    # mtspr CTR,r12
    # bctrl # CSamusHud::DisplayHudMemo

    # end:
    # lwz r0,0x30(r1)
    # mtspr LR,r0
    # addi r1,r1,0x2C
    # blr

    message_receiver_patch = [
        0x94, 0x21, 0xFF, 0xD4,
        0x7C, 0x08, 0x02, 0xA6,
        0x90, 0x01, 0x00, 0x30,
        0x88, 0x83, 0x00, 0x02,
        0x2C, 0x04, 0x00, 0x00,
        0x41, 0x82, 0x00, 0x60,
        0x38, 0xC0, 0x00, 0x00,
        0x98, 0xC3, 0x00, 0x02,
        0x3C, 0xA0, 0x41, 0x00,
        0x38, 0xE0, 0x00, 0x01,
        0x39, 0x20, 0x00, 0x09,
        0x90, 0xA1, 0x00, 0x10,
        0x98, 0xE1, 0x00, 0x14,
        0x98, 0xC1, 0x00, 0x15,
        0x98, 0xC1, 0x00, 0x16,
        0x98, 0xE1, 0x00, 0x17,
        0x91, 0x21, 0x00, 0x18,
        0x38, 0x61, 0x00, 0x1C,
        0x3C, 0x80, message_receiver_string_ref[0], message_receiver_string_ref[1],
        0x60, 0x84, message_receiver_string_ref[2], message_receiver_string_ref[3],
        0x3D, 0x80, address_wstring_constructor[0], address_wstring_constructor[1],
        0x61, 0x8C, address_wstring_constructor[2], address_wstring_constructor[3],
        0x7D, 0x89, 0x03, 0xA6,
        0x4E, 0x80, 0x04, 0x21,
        0x38, 0x81, 0x00, 0x10,
        0x3D, 0x80, address_display_hud_memo[0], address_display_hud_memo[1],
        0x61, 0x8C, address_display_hud_memo[2], address_display_hud_memo[3],
        0x7D, 0x89, 0x03, 0xA6,
        0x4E, 0x80, 0x04, 0x21,
        0x80, 0x01, 0x00, 0x30,
        0x7C, 0x08, 0x03, 0xA6,
        0x38, 0x21, 0x00, 0x2C,
        0x4E, 0x80, 0x00, 0x20,
    ]

    dol_file.write(patch_addresses.update_hint_state, message_receiver_patch)


def _apply_game_options_patch(game_options_constructor_offset: int, user_preferences: EchoesUserPreferences,
                              dol_file: DolFile):
    patch = [
        # Unknown purpose, but keep for safety
        0x93, 0xe1, 0x00, 0x1c,  # *(r1 + 0x1c) = r31   (stw r31,0x1c(r1))

        0x7c, 0x7f, 0x1b, 0x78,  # r31 = r3 (ori r31,r3,r3)
        0x38, 0x61, 0x00, 0x08,  # r3 = r1 + 0x8 (addi r3,r1,0x8) For a later function call we don't touch
    ]

    for i, preference_name in enumerate(_PREFERENCES_ORDER):
        value = getattr(user_preferences, preference_name)
        if isinstance(value, Enum):
            value = value.value
        encoded_value = struct.pack(">h", value)
        patch.extend([
            0x38, 0x00, encoded_value[0], encoded_value[1],  # r0 = value (li r0, value)
            0x90, 0x1f, 0x00, (0x04 * i),  # *(r31 + offset) = r0  (stw r0,offset(r31))
        ])

    flag_values = [
        getattr(user_preferences, flag_name)
        if flag_name is not None else False
        for flag_name in _FLAGS_ORDER
    ]
    bit_mask = int("".join(str(int(flag)) for flag in flag_values), 2)
    patch.extend([
        0x38, 0x00, 0x00, bit_mask,
        0x98, 0x1f, 0x00, (0x04 * len(_PREFERENCES_ORDER)),
    ])
    patch.extend([
        0x38, 0x00, 0x00, 0x00,  # li r0, value)
        0x90, 0x1f, 0x00, 0x2c,  # stw r0 ,0x2c (r31)
        0x90, 0x1f, 0x00, 0x30,  # stw r0 ,0x30 (r31)
        0x90, 0x1f, 0x00, 0x34,  # stw r0 ,0x34 (r31)
    ])

    # final_offset = 0x15E9E4
    # total_bytes_to_patch = final_offset - game_options_offset
    total_bytes_to_patch = 136
    bytes_to_fill = total_bytes_to_patch - len(patch)

    if bytes_to_fill < 0:
        raise RuntimeError(f"Our patch ({len(patch)}) is bigger than the space we have ({total_bytes_to_patch}).")

    if bytes_to_fill % 4 != 0:
        raise RuntimeError(f"The space left ({bytes_to_fill}) for is not a multiple of 4")

    # ori r0, r0, 0x0 is a no-op
    patch.extend([0x60, 0x00, 0x00, 0x00] * (bytes_to_fill // 4))
    dol_file.write(game_options_constructor_offset + 8 * 4, patch)


def _apply_energy_tank_capacity_patch(patch_addresses: HealthCapacityAddresses, game_specific: EchoesGameSpecific,
                                      dol_file: DolFile):
    """
    Patches the base health capacity and the energy tank capacity with matching values.
    """
    tank_capacity = game_specific.energy_per_tank

    dol_file.write(patch_addresses.base_health_capacity, struct.pack(">f", tank_capacity - 1))
    dol_file.write(patch_addresses.energy_tank_capacity, struct.pack(">f", tank_capacity))


def _apply_beam_cost_patch(patch_addresses: BeamCostAddresses, game_specific: EchoesGameSpecific,
                           dol_file: DolFile):
    uncharged_costs = []
    charged_costs = []
    combo_costs = []
    missile_costs = []
    ammo_types = []

    for beam_config in game_specific.beam_configurations:
        uncharged_costs.append(beam_config.uncharged_cost)
        charged_costs.append(beam_config.charged_cost)
        combo_costs.append(beam_config.combo_ammo_cost)
        missile_costs.append(beam_config.combo_missile_cost)
        ammo_types.append((
            beam_config.ammo_a.index if beam_config.ammo_a is not None else -1,
            beam_config.ammo_b.index if beam_config.ammo_b is not None else -1,
        ))

    uncharged_costs_patch = struct.pack(">llll", *uncharged_costs)
    charged_costs_patch = struct.pack(">llll", *charged_costs)
    combo_costs_patch = struct.pack(">llll", *combo_costs)
    missile_costs_patch = struct.pack(">llll", *missile_costs)

    def enc(i, x):
        return struct.pack(">h", ammo_types[i][x])

    # TODO: patch CPlayerGun::IsOutOfAmmoToShoot

    # The following patch also changes the fact that the game doesn't check if there's enough ammo for Power Beam
    # we start our patch right after the `addi r3,r31,0x0`
    ammo_type_patch_offset = 0x40
    ammo_type_patch = [
        0x81, 0x59, 0x07, 0x74,  # lwz r10,0x774(r25)           # r10 = get current beam
        0x55, 0x4a, 0x10, 0x3a,  # rlwinm r10,r10,0x2,0x0,0x1d  # r10 *= 4

        0x7c, 0x03, 0x50, 0x2e,  # lwzx r0,r3,r10               # r0 = BeamIdToUnchargedShotAmmoCost[currentBeam]
        0x90, 0x1d, 0x00, 0x00,  # stw r0,0x0(r29)              # *outBeamAmmoCost = r0

        0x81, 0x59, 0x07, 0x74,  # lwz r10,0x774(r25)           # r10 = get current beam
        0x39, 0x4a, 0x00, 0x01,  # addi r10,r10,0x1             # r10 = r10
        0x7d, 0x49, 0x03, 0xa6,  # mtspr CTR,r10                # count_register = r10

        # Power Beam
        0x42, 0x00, 0x00, 0x10,  # bdnz dark_beam               # if (--count_register > 0) goto
        0x38, 0x60, *enc(0, 0),  # li r3, <type>
        0x39, 0x20, *enc(0, 1),  # li r9, <type>
        0x42, 0x80, 0x00, 0x2c,  # b update_out_beam_type

        # Dark Beam
        0x42, 0x00, 0x00, 0x10,  # bdnz dark_beam               # if (--count_register > 0) goto
        0x38, 0x60, *enc(1, 0),  # li r3, <type>
        0x39, 0x20, *enc(1, 1),  # li r9, <type>
        0x42, 0x80, 0x00, 0x1c,  # b update_out_beam_type

        # Light Beam
        0x42, 0x00, 0x00, 0x10,  # bdnz light_beam               # if (--count_register > 0) goto
        0x38, 0x60, *enc(2, 0),  # li r3, <type>
        0x39, 0x20, *enc(2, 1),  # li r9, <type>
        0x42, 0x80, 0x00, 0x0c,  # b update_out_beam_type

        # Annihilator Beam
        0x38, 0x60, *enc(3, 0),  # li r3, <type>
        0x39, 0x20, *enc(3, 1),  # li r9, <type>

        # update_out_beam_type
        0x90, 0x7b, 0x00, 0x00,  # stw r0,0x0(r27)              # *outBeamAmmoTypeA = r3
        0x91, 0x3c, 0x00, 0x00,  # stw r0,0x0(r28)              # *outBeamAmmoTypeB = r8

        0x42, 0x80, 0x00, 0x18,  # b body_end
        # jump to the code for getting the charged/combo costs and then check if has ammo
        # The address in question is at 0x801ccd64 for NTSC
    ]

    dol_file.write(patch_addresses.uncharged_cost, uncharged_costs_patch)
    dol_file.write(patch_addresses.charged_cost, charged_costs_patch)
    dol_file.write(patch_addresses.charge_combo_ammo_cost, combo_costs_patch)
    dol_file.write(patch_addresses.charge_combo_missile_cost, missile_costs_patch)
    dol_file.write(patch_addresses.get_beam_ammo_type_and_costs + ammo_type_patch_offset, ammo_type_patch)


def _apply_safe_zone_heal_patch(patch_addresses: SafeZoneAddresses,
                                sda2_base: int,
                                game_specific: EchoesGameSpecific,
                                dol_file: DolFile):
    heal_per_second = game_specific.safe_zone_heal_per_second
    offset = patch_addresses.heal_per_frame_constant - sda2_base

    dol_file.write(patch_addresses.heal_per_frame_constant, struct.pack(">f", heal_per_second / 60))
    dol_file.write(patch_addresses.increment_health_fmr, [
        0xc0, 0x22, *struct.pack(">h", offset),  # lfs f1, <offset>(r2)
    ])


def apply_patches(game_root: Path, game_specific: EchoesGameSpecific, user_preferences: EchoesUserPreferences):
    dol_file = DolFile(_get_dol_path(game_root))

    version_patches = _read_binary_version(dol_file)

    dol_file.set_editable(True)
    with dol_file:
        _apply_string_display_patch(version_patches.string_display, dol_file)
        _apply_game_options_patch(version_patches.game_options_constructor_address,
                                  user_preferences, dol_file)
        _apply_energy_tank_capacity_patch(version_patches.health_capacity, game_specific, dol_file)
        _apply_beam_cost_patch(version_patches.beam_cost_addresses, game_specific, dol_file)
        _apply_safe_zone_heal_patch(version_patches.safe_zone, version_patches.sda2_base, game_specific, dol_file)


def _get_dol_path(game_root: Path) -> Path:
    return game_root.joinpath("sys/main.dol")


def _read_binary_version(dol_file: DolFile) -> PatchesForVersion:
    dol_file.set_editable(False)
    with dol_file:
        for version in ALL_VERSIONS_PATCHES:
            build_string = dol_file.read(version.build_string_address, len(version.build_string))
            if build_string == version.build_string:
                return version

    raise RuntimeError(f"Unsupported game version")



import multiprocessing
import traceback
from typing import Callable, Union

from randovania.generator import generator
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.resolver import debug


def _generate_layout_worker(output_pipe,
                            permalink: Permalink,
                            validate_after_generation: bool,
                            timeout_during_generation: bool,
                            debug_level: int):
    try:
        def status_update(message: str):
            output_pipe.send(message)

        debug.set_level(debug_level)
        extra_args = {}
        if not timeout_during_generation:
            extra_args["timeout"] = None

        layout_description = generator.generate_description(permalink,
                                                            status_update=status_update,
                                                            validate_after_generation=validate_after_generation,
                                                            **extra_args)
        output_pipe.send(layout_description)
    except Exception as e:
        traceback.print_exc()
        output_pipe.send(e)


def generate_layout(permalink: Permalink,
                    status_update: Callable[[str], None],
                    validate_after_generation: bool,
                    timeout_during_generation: bool,
                    ) -> LayoutDescription:
    receiving_pipe, output_pipe = multiprocessing.Pipe(False)

    debug_level = debug.debug_level()
    if not permalink.spoiler:
        debug_level = 0

    process = multiprocessing.Process(
        target=_generate_layout_worker,
        args=(output_pipe, permalink, validate_after_generation, timeout_during_generation, debug_level)
    )
    process.start()
    try:
        result: Union[Exception, LayoutDescription] = None
        while result is None:
            pipe_input = receiving_pipe.recv()
            if isinstance(pipe_input, (LayoutDescription, Exception)):
                result = pipe_input
            else:
                status_update(pipe_input)
    finally:
        process.terminate()

    if isinstance(result, Exception):
        raise result
    else:
        return result



from pathlib import Path
from unittest.mock import patch, MagicMock, ANY

import pytest

from randovania.games.prime import dol_patcher
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences


@patch("randovania.games.prime.dol_patcher._apply_beam_cost_patch", autospec=True)
@patch("randovania.games.prime.dol_patcher._apply_energy_tank_capacity_patch", autospec=True)
@patch("randovania.games.prime.dol_patcher._apply_game_options_patch", autospec=True)
@patch("randovania.games.prime.dol_patcher._apply_string_display_patch", autospec=True)
@patch("randovania.games.prime.dol_patcher.DolFile")
@patch("randovania.games.prime.dol_patcher._get_dol_path", autospec=True)
@patch("randovania.games.prime.dol_patcher._read_binary_version", autospec=True)
def test_apply_patches(mock_read_binary_version: MagicMock,
                       mock_get_dol_path: MagicMock,
                       mock_dol_file_constructor: MagicMock,
                       mock_apply_string_display_patch: MagicMock,
                       mock_apply_game_options_patch: MagicMock,
                       mock_apply_energy_tank_capacity_patch: MagicMock,
                       mock_apply_beam_cost_patch: MagicMock,
                       ):
    # Setup
    game_root = MagicMock()
    game_patches = MagicMock()
    user_preferences = MagicMock()
    version_patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    mock_read_binary_version.return_value = version_patches
    dol_file = mock_dol_file_constructor.return_value

    # Run
    dol_patcher.apply_patches(game_root, game_patches.game_specific, user_preferences)

    # Assert
    mock_read_binary_version.assert_called_once_with(dol_file)
    mock_get_dol_path.assert_called_once_with(game_root)
    mock_dol_file_constructor.assert_called_once_with(mock_get_dol_path.return_value)
    mock_apply_string_display_patch.assert_called_once_with(version_patches.string_display, dol_file)
    mock_apply_game_options_patch.assert_called_once_with(
        version_patches.game_options_constructor_address,
        user_preferences, dol_file
    )
    mock_apply_energy_tank_capacity_patch.assert_called_once_with(
        version_patches.health_capacity,
        game_patches.game_specific, dol_file
    )
    mock_apply_beam_cost_patch.assert_called_once_with(
        version_patches.beam_cost_addresses,
        game_patches.game_specific, dol_file
    )


def test_get_dol_path():
    assert dol_patcher._get_dol_path(Path("foo")) == Path("foo", "sys", "main.dol")


@pytest.mark.parametrize("version", dol_patcher.ALL_VERSIONS_PATCHES)
def test_read_binary_version(version):
    dol_file = MagicMock()
    dol_file.read.return_value = version.build_string

    # Run
    result = dol_patcher._read_binary_version(dol_file)

    # Assert
    dol_file.set_editable.assert_called_once_with(False)
    assert result == version


def test_apply_string_display_patch():
    offsets = dol_patcher.StringDisplayPatchAddresses(0x15, 0, 0, 0, 0, 0)
    dol_file = MagicMock()

    # Run
    dol_patcher._apply_string_display_patch(offsets, dol_file)

    # Assert
    dol_file.write.assert_called_once_with(0x15, ANY)


def test_apply_game_options_patch():
    offset = 0x15
    dol_file = MagicMock()
    user_preferences = EchoesUserPreferences()

    # Run
    dol_patcher._apply_game_options_patch(offset, user_preferences, dol_file)

    # Assert
    dol_file.write.assert_called_once_with(53, ANY)



import copy
import dataclasses
import logging
import struct
from typing import Optional, List, Dict

from randovania.game_connection.backend_choice import GameBackendChoice
from randovania.game_connection.connection_base import ConnectionBase, InventoryItem, GameConnectionStatus
from randovania.game_description import data_reader
from randovania.game_description.game_description import GameDescription
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_info import CurrentResources, add_resource_gain_to_current_resources
from randovania.games.game import RandovaniaGame
from randovania.games.prime import dol_patcher, default_data
from randovania.games.prime.dol_patcher import PatchesForVersion


@dataclasses.dataclass(frozen=True)
class MemoryOperation:
    address: int
    offset: Optional[int] = None
    read_byte_count: Optional[int] = None
    write_bytes: Optional[bytes] = None

    @property
    def byte_count(self) -> int:
        if self.read_byte_count is not None:
            return self.read_byte_count
        if self.write_bytes is not None:
            return len(self.write_bytes)
        return 0

    def validate_byte_sizes(self):
        if (self.write_bytes is not None
                and self.read_byte_count is not None
                and len(self.write_bytes) != self.read_byte_count):
            raise ValueError(f"Attempting to read {self.read_byte_count} bytes while writing {len(self.write_bytes)}.")

    def __str__(self):
        address_text = f"0x{self.address:08x}"
        if self.offset is not None:
            address_text = f"*{address_text} {self.offset:+05x}"

        operation_pretty = []
        if self.read_byte_count is not None:
            operation_pretty.append(f"read {self.read_byte_count} bytes")
        if self.write_bytes is not None:
            operation_pretty.append(f"write {self.write_bytes.hex()}")

        return f"At {address_text}, {' and '.join(operation_pretty)}"


def _powerup_offset(item_index: int) -> int:
    powerups_offset = 0x58
    vector_data_offset = 0x4
    powerup_size = 0xc
    return (powerups_offset + vector_data_offset) + (item_index * powerup_size)


def _add_pickup_to_resources(pickup: PickupEntry, inventory: CurrentResources) -> CurrentResources:
    return add_resource_gain_to_current_resources(
        pickup.resource_gain(inventory),
        copy.copy(inventory)
    )


def _capacity_for(item: ItemResourceInfo,
                  changed_items: Dict[ItemResourceInfo, InventoryItem],
                  inventory: Dict[ItemResourceInfo, InventoryItem]):
    if item in changed_items:
        return changed_items[item].capacity
    elif item in inventory:
        return inventory[item].capacity
    else:
        return 0


class ConnectionBackend(ConnectionBase):
    patches: Optional[PatchesForVersion] = None
    _checking_for_collected_index: bool = False
    _games: Dict[RandovaniaGame, GameDescription]
    _inventory: Dict[ItemResourceInfo, InventoryItem]
    _enabled: bool = True

    # Messages
    message_queue: List[str]
    message_cooldown: float = 0.0
    _last_message_size: int = 0

    # Multiworld
    _pickups_to_give: List[PickupEntry]
    _permanent_pickups: List[PickupEntry]

    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)

        self._games = {}
        self._inventory = {}
        self.message_queue = []
        self._pickups_to_give = []
        self._permanent_pickups = []

    @property
    def current_status(self) -> GameConnectionStatus:
        raise NotImplementedError()

    @property
    def backend_choice(self) -> GameBackendChoice:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        raise NotImplementedError()

    async def update(self, dt: float):
        raise NotImplementedError()

    @property
    def lock_identifier(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def checking_for_collected_index(self):
        return self._checking_for_collected_index

    @checking_for_collected_index.setter
    def checking_for_collected_index(self, value):
        self._checking_for_collected_index = value

    @ConnectionBase.tracking_inventory.setter
    def tracking_inventory(self, value: bool):
        self._tracking_inventory = value

    @ConnectionBase.displaying_messages.setter
    def displaying_messages(self, value: bool):
        self._displaying_messages = value

    def set_connection_enabled(self, value: bool):
        self._enabled = value
        if not value:
            self.patches = None

    # Game Backend Stuff
    async def _perform_memory_operations(self, ops: List[MemoryOperation]) -> List[Optional[bytes]]:
        raise NotImplementedError()

    async def _perform_single_memory_operations(self, op: MemoryOperation) -> Optional[bytes]:
        result = await self._perform_memory_operations([op])
        return result[0]

    @property
    def game(self) -> GameDescription:
        game_enum = self.patches.game
        if game_enum not in self._games:
            self._games[game_enum] = data_reader.decode_data(default_data.read_json_then_binary(game_enum)[1])
        return self._games[game_enum]

    async def _identify_game(self) -> bool:
        if self.patches is not None:
            return True

        for version in dol_patcher.ALL_VERSIONS_PATCHES:
            try:
                # We're reading these build strings separately because combining would go above the maximum size allowed
                # for read operations
                operation = MemoryOperation(version.build_string_address, read_byte_count=len(version.build_string))
                build_string = await self._perform_single_memory_operations(operation)
            except RuntimeError:
                return False

            if build_string == version.build_string:
                self.patches = version
                self.logger.info(f"_identify_game: identified game as {version.description}")
                return True

        return False

    async def _update_current_world(self):
        try:
            world_asset_id = await self._perform_single_memory_operations(
                MemoryOperation(self.patches.game_state_pointer,
                                offset=4,
                                read_byte_count=4)
            )
            if world_asset_id is None:
                raise KeyError()
            self._world = self.game.world_list.world_by_asset_id(struct.unpack(">I", world_asset_id)[0])

        except (KeyError, RuntimeError):
            self._world = None

    def _get_player_state_pointer(self) -> int:
        cstate_manager = self.patches.string_display.cstate_manager_global
        return cstate_manager + 0x150c

    def get_current_inventory(self) -> Dict[ItemResourceInfo, InventoryItem]:
        return self._inventory

    async def _read_item(self, item: ItemResourceInfo) -> InventoryItem:
        player_state_pointer = self._get_player_state_pointer()

        op_result = await self._perform_single_memory_operations(MemoryOperation(
            address=player_state_pointer,
            read_byte_count=8,
            offset=_powerup_offset(item.index),
        ))

        return InventoryItem(*struct.unpack(">II", op_result))

    async def _get_inventory(self) -> Dict[ItemResourceInfo, InventoryItem]:
        player_state_pointer = self._get_player_state_pointer()

        memory_ops = [
            MemoryOperation(
                address=player_state_pointer,
                read_byte_count=8,
                offset=_powerup_offset(item.index),
            )
            for item in self.game.resource_database.item
        ]

        ops_result = await self._perform_memory_operations(memory_ops)

        inventory = {}
        for item, memory_result in zip(self.game.resource_database.item, ops_result):
            inventory[item] = InventoryItem(*struct.unpack(">II", memory_result))

        return inventory

    async def _perform_write_inventory(self, changed_items: Dict[ItemResourceInfo, InventoryItem]):
        player_state_pointer = self._get_player_state_pointer()

        memory_ops = []
        for item in self.game.resource_database.item:
            if item in changed_items:
                memory_ops.append(MemoryOperation(
                    address=player_state_pointer,
                    write_bytes=struct.pack(">II", *changed_items[item]),
                    read_byte_count=8,
                    offset=_powerup_offset(item.index),
                ))
                self.logger.debug(f"Setting {item.long_name} to "
                                  f"{changed_items[item].amount}/{changed_items[item].capacity} "
                                  f"({memory_ops[-1].write_bytes.hex()})")

        # Suit Model
        dark_suit = self.game.resource_database.get_item(13)
        light_suit = self.game.resource_database.get_item(14)
        if dark_suit in changed_items or light_suit in changed_items:
            if _capacity_for(light_suit, changed_items, self._inventory) > 0:
                new_suit = 2
            elif _capacity_for(dark_suit, changed_items, self._inventory) > 0:
                new_suit = 1
            else:
                new_suit = 0
            memory_ops.append(MemoryOperation(
                address=player_state_pointer,
                write_bytes=struct.pack(">I", new_suit),
                offset=84,
            ))
            self.logger.debug(f"Setting suit to {new_suit}. ({memory_ops[-1].write_bytes.hex()})")

        energy_tank = self.game.resource_database.energy_tank
        if energy_tank in changed_items:
            health_data = await self._perform_memory_operations([
                MemoryOperation(player_state_pointer, read_byte_count=4, offset=20),
                MemoryOperation(self.patches.health_capacity.base_health_capacity, read_byte_count=4),
                MemoryOperation(self.patches.health_capacity.energy_tank_capacity, read_byte_count=4),
            ])
            current_health, base_health_capacity, energy_tank_capacity = struct.unpack(">fff", b"".join(health_data))
            new_health = changed_items[energy_tank].amount * energy_tank_capacity + base_health_capacity
            if changed_items[energy_tank] < changed_items[energy_tank]:
                new_health = min(new_health, current_health)

            memory_ops.append(MemoryOperation(
                address=player_state_pointer,
                write_bytes=struct.pack(">f", new_health),
                offset=20,
            ))
            self.logger.debug(f"Setting health to {new_health}. ({memory_ops[-1].write_bytes.hex()})")

        # FIXME: check if the value read is what we expected, and then re-writes if needed
        result = await self._perform_memory_operations(memory_ops)
        return changed_items

    async def _write_item(self, item: ItemResourceInfo, value: InventoryItem):
        await self._perform_write_inventory({item: value})

    async def _update_inventory(self, new_inventory: Dict[ItemResourceInfo, InventoryItem]):
        current_inventory = self.get_current_inventory()

        changed_items = {item: new_inventory[item]
                         for item in self.game.resource_database.item
                         if new_inventory[item] != current_inventory[item]}
        if changed_items:
            new_inventory.update(await self._perform_write_inventory(changed_items))
            self._inventory = new_inventory

    # Multiworld
    async def _check_for_collected_index(self):
        multiworld_magic_item = self.game.resource_database.multiworld_magic_item
        item_percentage = self.game.resource_database.item_percentage

        new_magic_item = None
        current_inventory = None
        new_inventory = None

        if self._tracking_inventory:
            current_inventory = self.get_current_inventory()
            new_inventory = copy.copy(current_inventory)
            magic_inv = current_inventory[multiworld_magic_item]
        else:
            magic_inv = await self._read_item(multiworld_magic_item)

        magic_capacity = magic_inv.capacity
        if magic_inv.amount > 0:
            self.logger.info(f"_check_for_collected_index: magic item was at {magic_inv.amount}/{magic_capacity}")
            magic_capacity -= magic_inv.amount
            new_magic_item = InventoryItem(0, magic_inv.capacity - magic_inv.amount)
            await self._emit_location_collected(magic_inv.amount - 1)

        if self._pickups_to_give or magic_capacity < len(self._permanent_pickups):
            self.logger.info(f"_check_for_collected_index: {len(self._pickups_to_give)} pickups to give, "
                             f"{len(self._permanent_pickups)} permanent pickups, magic {magic_capacity}")

            if current_inventory is None:
                current_inventory = await self._get_inventory()
                self._inventory = current_inventory
                new_inventory = copy.copy(current_inventory)

            inventory_resources: CurrentResources = {
                item: inv_item.capacity
                for item, inv_item in current_inventory.items()
            }

            while self._pickups_to_give:
                inventory_resources = _add_pickup_to_resources(self._pickups_to_give.pop(), inventory_resources)

            for pickup in self._permanent_pickups[magic_capacity:]:
                inventory_resources = _add_pickup_to_resources(pickup, inventory_resources)

            for item, new_capacity in inventory_resources.items():
                if item == item_percentage:
                    continue
                old = new_inventory[item]
                delta = new_capacity - old.capacity
                new_amount = old.amount
                if delta > 0:
                    new_amount += delta
                new_inventory[item] = InventoryItem(min(new_amount, new_capacity, item.max_capacity),
                                                    min(new_capacity, item.max_capacity))

            new_magic_item = InventoryItem(0, len(self._permanent_pickups))

        if new_magic_item is not None:
            if new_inventory is not None:
                new_inventory[multiworld_magic_item] = new_magic_item
                await self._update_inventory(new_inventory)
            else:
                await self._write_item(multiworld_magic_item, new_magic_item)

    def send_pickup(self, pickup: PickupEntry):
        self._pickups_to_give.append(pickup)

    def set_permanent_pickups(self, pickups: List[PickupEntry]):
        self._permanent_pickups = pickups

    # Display Message
    def display_message(self, message: str):
        if self._displaying_messages:
            self.logger.info(f"Queueing message '{message}'. "
                             f"Queue has {len(self.message_queue)} elements and "
                             f"current cooldown is {self.message_cooldown}")
            self.message_queue.append(message)
        else:
            self.logger.info(f"Ignoring message '{message}', _displaying_messages is False.")

    async def _send_message_from_queue(self, dt: float):
        # If there's no messages, don't bother
        if not self.message_queue:
            return

        has_message_address = self.patches.string_display.cstate_manager_global + 0x2

        # There's already a message pending, stop
        has_message = await self._perform_single_memory_operations(MemoryOperation(has_message_address,
                                                                                   read_byte_count=1))

        if has_message != b"\x00":
            self.logger.info("_send_message_from_queue: game already has a pending message")
            return

        self.message_cooldown = max(self.message_cooldown - dt, 0.0)

        # There's a cooldown for next message!
        if self.message_cooldown > 0:
            self.logger.debug(f"_send_message_from_queue: current cooldown is {self.message_cooldown}")
            return

        message = self.message_queue.pop(0)
        overhead_size = 6  # 2 bytes for an extra char to differentiate sizes
        encoded_message = message.encode("utf-16_be")[:self.patches.string_display.max_message_size - overhead_size]

        # The game doesn't handle very well a string at the same address with same size being
        # displayed multiple times
        if len(encoded_message) == self._last_message_size:
            encoded_message += b'\x00 '
        self._last_message_size = len(encoded_message)

        # Add the null terminator
        encoded_message += b"\x00\x00"
        if len(encoded_message) & 3:
            # Ensure the size is a multiple of 4
            num_to_align = (len(encoded_message) | 3) - len(encoded_message) + 1
            encoded_message += b"\x00" * num_to_align

        await self._perform_memory_operations([
            # The message string
            MemoryOperation(self.patches.string_display.message_receiver_string_ref,
                            write_bytes=encoded_message),

            # Notify game to display message
            MemoryOperation(has_message_address, write_bytes=b"\x01"),
        ])
        self.message_cooldown = 4

        self.logger.info(f"_send_message_from_queue: sent '{message}' to game. "
                         f"{len(self.message_queue)} messages left.")

    async def update_current_inventory(self):
        self._inventory = await self._get_inventory()

    async def _interact_with_game(self, dt):
        await self._update_current_world()
        if self._world is not None:
            await self._send_message_from_queue(dt)
            if self._tracking_inventory:
                await self.update_current_inventory()

            if self.checking_for_collected_index:
                await self._check_for_collected_index()



import asyncio
import datetime
import locale
import logging.config
import os
import sys
import typing
from argparse import ArgumentParser
from pathlib import Path

from PySide2 import QtCore
from PySide2.QtWidgets import QApplication, QMessageBox
from asyncqt import asyncClose

from randovania.gui.lib import theme, common_qt_lib


def catch_exceptions(t, val, tb):
    if not isinstance(val, KeyboardInterrupt):
        QMessageBox.critical(None,
                             "An exception was raised",
                             "An unhandled Exception occurred:\n{}".format(val))
    old_hook(t, val, tb)


old_hook = sys.excepthook


def catch_exceptions_async(loop, context):
    if 'future' in context:
        future: asyncio.Future = context['future']
        logging.exception(context["message"], exc_info=future.exception())
    else:
        logging.critical(str(context))


async def show_main_window(app: QApplication, options, is_preview: bool):
    from randovania.interface_common.preset_manager import PresetManager
    preset_manager = PresetManager(options.data_dir)

    await preset_manager.load_user_presets()

    from randovania.gui.main_window import MainWindow
    main_window = MainWindow(options, preset_manager, app.network_client, is_preview)
    app.main_window = main_window
    main_window.show()
    await main_window.request_new_data()


def show_tracker(app: QApplication):
    from randovania.gui.auto_tracker_window import AutoTrackerWindow

    app.tracker = AutoTrackerWindow(app.game_connection)
    app.tracker.show()


def show_game_details(app: QApplication, options, game: Path):
    from randovania.layout.layout_description import LayoutDescription
    from randovania.gui.seed_details_window import SeedDetailsWindow

    layout = LayoutDescription.from_file(game)
    details_window = SeedDetailsWindow(None, options)
    details_window.update_layout_description(layout)
    details_window.show()
    app.details_window = details_window


async def display_window_for(app, options, command: str, args):
    if command == "tracker":
        show_tracker(app)
    elif command == "main":
        await show_main_window(app, options, args.preview)
    elif command == "game":
        show_game_details(app, options, args.rdvgame)
    else:
        raise RuntimeError(f"Unknown command: {command}")


def create_backend(debug_game_backend: bool, options):
    from randovania.interface_common.options import Options
    options = typing.cast(Options, options)

    if debug_game_backend:
        from randovania.gui.debug_backend_window import DebugBackendWindow
        backend = DebugBackendWindow()
        backend.show()
    else:
        try:
            from randovania.game_connection.dolphin_backend import DolphinBackend
        except ImportError:
            common_qt_lib.show_install_visual_cpp_redist()
            raise SystemExit(1)

        from randovania.game_connection.nintendont_backend import NintendontBackend
        from randovania.game_connection.backend_choice import GameBackendChoice

        if options.game_backend == GameBackendChoice.NINTENDONT and options.nintendont_ip is not None:
            backend = NintendontBackend(options.nintendont_ip)
        else:
            backend = DolphinBackend()

    return backend


def _load_options():
    from randovania.interface_common.options import Options
    from randovania.gui.lib import startup_tools

    options = Options.with_default_data_dir()
    if not startup_tools.load_options_from_disk(options):
        raise SystemExit(1)

    theme.set_dark_theme(options.dark_mode)

    from randovania.layout.preset_migration import VersionedPreset
    for old_preset in options.data_dir.joinpath("presets").glob("*.randovania_preset"):
        old_preset.rename(old_preset.with_name(f"{old_preset.stem}.{VersionedPreset.file_extension()}"))

    return options


def start_logger(data_dir: Path, is_preview: bool):
    # Ensure the log dir exists early on
    log_dir = data_dir.joinpath("logs")
    log_dir.mkdir(parents=True, exist_ok=True)

    today = datetime.datetime.now().strftime("%Y-%m-%d")

    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            }
        },
        'handlers': {
            'default': {
                'level': 'DEBUG' if is_preview else 'INFO',
                'formatter': 'default',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout',  # Default is stderr
            },
            'local_app_data': {
                'level': 'DEBUG',
                'formatter': 'default',
                'class': 'logging.FileHandler',
                'filename': log_dir.joinpath(f"{today}.log"),
                'encoding': 'utf-8',
            }
        },
        'root': {
            'level': 'WARNING',
            'handlers': ['default', 'local_app_data'],
        }
    })


def create_loop(app: QApplication) -> asyncio.AbstractEventLoop:
    os.environ['QT_API'] = "PySide2"
    import asyncqt
    loop: asyncio.AbstractEventLoop = asyncqt.QEventLoop(app)
    asyncio.set_event_loop(loop)

    sys.excepthook = catch_exceptions
    loop.set_exception_handler(catch_exceptions_async)
    return loop


async def qt_main(app: QApplication, data_dir: Path, args):
    from randovania.gui.lib.qt_network_client import QtNetworkClient
    from randovania.game_connection.game_connection import GameConnection

    app.network_client = QtNetworkClient(data_dir)
    options = _load_options()

    backend = create_backend(args.debug_game_backend, options)
    app.game_connection = GameConnection(backend)
    app.game_connection.tracking_inventory = options.tracking_inventory
    app.game_connection.displaying_messages = options.displaying_messages

    @asyncClose
    async def _on_last_window_closed():
        await app.network_client.disconnect_from_server()
        await app.game_connection.stop()

    app.lastWindowClosed.connect(_on_last_window_closed, QtCore.Qt.QueuedConnection)

    await asyncio.gather(app.game_connection.start(),
                         display_window_for(app, options, args.command, args))


def run(args):
    locale.setlocale(locale.LC_ALL, "")  # use system's default locale
    QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

    data_dir = args.custom_network_storage
    if data_dir is None:
        from randovania.interface_common import persistence
        data_dir = persistence.user_data_dir()

    is_preview = args.preview
    start_logger(data_dir, is_preview)
    app = QApplication(sys.argv)

    loop = create_loop(app)
    with loop:
        loop.create_task(qt_main(app, data_dir, args))
        loop.run_forever()


def create_subparsers(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "gui",
        help="Run the Graphical User Interface"
    )
    parser.add_argument("--preview", action="store_true", help="Activates preview features")
    parser.add_argument("--custom-network-storage", type=Path, help="Use a custom path to store the network login.")
    parser.add_argument("--debug-game-backend", action="store_true", help="Opens the debug game backend.")

    gui_parsers = parser.add_subparsers(dest="command")
    gui_parsers.add_parser("main", help="Displays the Main Window").set_defaults(func=run)
    gui_parsers.add_parser("tracker", help="Opens only the auto tracker").set_defaults(func=run)

    game_parser = gui_parsers.add_parser("game", help="Opens an rdvgame")
    game_parser.add_argument("rdvgame", type=Path, help="Path ")
    game_parser.set_defaults(func=run)

    def check_command(args):
        if args.command is None:
            parser.print_help()
            raise SystemExit(1)

    parser.set_defaults(func=check_command)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'item_configuration_popup.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_ItemConfigurationPopup(object):
    def setupUi(self, ItemConfigurationPopup):
        if not ItemConfigurationPopup.objectName():
            ItemConfigurationPopup.setObjectName(u"ItemConfigurationPopup")
        ItemConfigurationPopup.resize(326, 197)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ItemConfigurationPopup.sizePolicy().hasHeightForWidth())
        ItemConfigurationPopup.setSizePolicy(sizePolicy)
        ItemConfigurationPopup.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(ItemConfigurationPopup)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.warning_label = QLabel(ItemConfigurationPopup)
        self.warning_label.setObjectName(u"warning_label")
        self.warning_label.setWordWrap(True)

        self.verticalLayout.addWidget(self.warning_label)

        self.included_box = QGroupBox(ItemConfigurationPopup)
        self.included_box.setObjectName(u"included_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.included_box.sizePolicy().hasHeightForWidth())
        self.included_box.setSizePolicy(sizePolicy1)
        self.included_box.setCheckable(True)
        self.gridLayout = QGridLayout(self.included_box)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.shuffled_spinbox = QSpinBox(self.included_box)
        self.shuffled_spinbox.setObjectName(u"shuffled_spinbox")
        self.shuffled_spinbox.setMinimum(1)
        self.shuffled_spinbox.setMaximum(99)

        self.gridLayout.addWidget(self.shuffled_spinbox, 2, 1, 1, 1)

        self.shuffled_radio = QRadioButton(self.included_box)
        self.shuffled_radio.setObjectName(u"shuffled_radio")

        self.gridLayout.addWidget(self.shuffled_radio, 2, 0, 1, 1)

        self.starting_radio = QRadioButton(self.included_box)
        self.starting_radio.setObjectName(u"starting_radio")

        self.gridLayout.addWidget(self.starting_radio, 1, 0, 1, 1)

        self.vanilla_radio = QRadioButton(self.included_box)
        self.vanilla_radio.setObjectName(u"vanilla_radio")

        self.gridLayout.addWidget(self.vanilla_radio, 0, 0, 1, 1)

        self.provided_ammo_label = QLabel(self.included_box)
        self.provided_ammo_label.setObjectName(u"provided_ammo_label")
        self.provided_ammo_label.setWordWrap(True)

        self.gridLayout.addWidget(self.provided_ammo_label, 3, 0, 1, 1)

        self.provided_ammo_spinbox = QSpinBox(self.included_box)
        self.provided_ammo_spinbox.setObjectName(u"provided_ammo_spinbox")

        self.gridLayout.addWidget(self.provided_ammo_spinbox, 3, 1, 1, 1)


        self.verticalLayout.addWidget(self.included_box)

        self.button_box = QDialogButtonBox(ItemConfigurationPopup)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Save)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(ItemConfigurationPopup)

        QMetaObject.connectSlotsByName(ItemConfigurationPopup)
    # setupUi

    def retranslateUi(self, ItemConfigurationPopup):
        ItemConfigurationPopup.setWindowTitle(QCoreApplication.translate("ItemConfigurationPopup", u"Item Configuration", None))
        self.warning_label.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Hidden warning!", None))
        self.included_box.setTitle(QCoreApplication.translate("ItemConfigurationPopup", u"Included", None))
        self.shuffled_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Shuffled", None))
        self.starting_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Starting Item", None))
        self.vanilla_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Vanilla", None))
#if QT_CONFIG(tooltip)
        self.provided_ammo_label.setToolTip(QCoreApplication.translate("ItemConfigurationPopup", u"<html><head/><body><p>When this item is collected, it also gives this amount of the given ammos.</p><p>This is included in the calculation of how much each pickup of this ammo gives.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.provided_ammo_label.setText(QCoreApplication.translate("ItemConfigurationPopup", u"<html><head/><body><p>Provided Ammo<br/>(XXXX and YYYY)</p></body></html>", None))
    # retranslateUi




import asyncio
import atexit
import copy
import logging
import os
import re
import subprocess
from pathlib import Path

_FOLDER = Path(__file__).parent
AGENT_OUTPUT_PATTERN = re.compile(
    r'SSH_AUTH_SOCK=(?P<socket>[^;]+).*SSH_AGENT_PID=(?P<pid>\d+)', re.MULTILINE | re.DOTALL)


def _kill_agent():
    logging.info('killing previously started ssh-agent')
    subprocess.run(['ssh-agent', '-k'])
    del os.environ['SSH_AUTH_SOCK']
    del os.environ['SSH_AGENT_PID']


def _setup_agent():
    process = subprocess.run(['ssh-agent', '-s'], stdout=subprocess.PIPE, universal_newlines=True)
    match = AGENT_OUTPUT_PATTERN.search(process.stdout)
    if match is None:
        raise Exception('Could not parse ssh-agent output. It was: {}'.format(process.stdout))
    agent_data = match.groupdict()
    logging.info('ssh agent data: {}'.format(agent_data))
    logging.info('exporting ssh agent environment variables')
    os.environ['SSH_AUTH_SOCK'] = agent_data['socket']
    os.environ['SSH_AGENT_PID'] = agent_data['pid']
    atexit.register(_kill_agent)


async def main():
    if os.environ["GITHUB_REF"].startswith("refs/tags/"):
        server_environment = "production"
    else:
        server_environment = "staging"

    remote_host = "randovania.metroidprime.run"
    ssh_key = os.environ["DEPLOYMENT_KEY"]

    _setup_agent()
    subprocess.run(['ssh-add', "-"], check=True, input=ssh_key, text=True)

    new_env = copy.copy(os.environ)
    new_env["DOCKER_HOST"] = f"ssh://root@{remote_host}"
    new_env["DOMAIN"] = remote_host
    if server_environment == "production":
        new_env["PATH_PREFIX"] = "randovania"
        new_env["DATA_PATH"] = "/var/randovania/production/data"
        new_env["VERSION"] = os.environ["GITHUB_REF"].split("/")[-1]

    elif server_environment == "staging":
        new_env["PATH_PREFIX"] = "randovania-staging"
        new_env["DATA_PATH"] = "/var/randovania/staging/data"
        new_env["VERSION"] = "sha-{}".format(os.environ["GITHUB_SHA"][:8])
    else:
        raise ValueError(f"Unknown server_environment: {server_environment}")

    subprocess.run(
        [
            "docker-compose",
            "-p", f"randovania-{server_environment}",
            "up",
            "-d",
        ],
        check=True,
        env=new_env,
        cwd=_FOLDER.joinpath("server-docker")
    )


if __name__ == '__main__':
    asyncio.run(main())



import copy
import struct

import pytest
from mock import AsyncMock, call

from randovania.game_connection import connection_backend
from randovania.game_connection.connection_backend import ConnectionBackend, MemoryOperation
from randovania.game_connection.connection_base import InventoryItem
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.games.prime import dol_patcher


@pytest.fixture(name="backend")
def dolphin_backend():
    backend = ConnectionBackend()
    backend._perform_memory_operations = AsyncMock()
    return backend


@pytest.mark.asyncio
async def test_identify_game_ntsc(backend):
    # Setup
    backend._perform_memory_operations.return_value = [b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32"]

    # Run
    assert await backend._identify_game()

    # Assert
    assert backend.patches is dol_patcher.ALL_VERSIONS_PATCHES[0]


@pytest.mark.asyncio
async def test_identify_game_error(backend):
    # Setup
    backend._perform_memory_operations.side_effect = RuntimeError("not connected")

    # Run
    assert not await backend._identify_game()


@pytest.mark.asyncio
async def test_identify_game_already_known(backend):
    # Setup
    backend.patches = True
    backend._perform_memory_operations.side_effect = RuntimeError("not connected")

    # Run
    assert await backend._identify_game()


@pytest.mark.parametrize(["message_original", "message_encoded", "previous_size"], [
    ("Magoo", b'\x00M\x00a\x00g\x00o\x00o\x00\x00', 0),
    ("Magoo2", b'\x00M\x00a\x00g\x00o\x00o\x002\x00\x00\x00\x00', 0),
    ("Magoo", b'\x00M\x00a\x00g\x00o\x00o\x00 \x00\x00\x00\x00', 10),
])
@pytest.mark.asyncio
async def test_send_message(backend, message_original, message_encoded, previous_size):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations.return_value = [b"\x00"]
    string_ref = backend.patches.string_display.message_receiver_string_ref
    has_message_address = backend.patches.string_display.cstate_manager_global + 0x2
    backend._last_message_size = previous_size

    # Run
    backend.display_message(message_original)
    await backend._send_message_from_queue(1)

    # Assert
    backend._perform_memory_operations.assert_has_awaits([
        call([MemoryOperation(has_message_address, read_byte_count=1)]),
        call([
            MemoryOperation(string_ref, write_bytes=message_encoded),
            MemoryOperation(has_message_address, write_bytes=b'\x01'),
        ]),
    ])


@pytest.mark.asyncio
async def test_send_message_from_queue_no_messages(backend):
    await backend._send_message_from_queue(1)
    backend._perform_memory_operations.assert_not_awaited()


@pytest.mark.asyncio
async def test_send_message_has_pending_message(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations.return_value = [b"\x01"]
    has_message_address = backend.patches.string_display.cstate_manager_global + 0x2

    # Run
    backend.display_message("Magoo")
    await backend._send_message_from_queue(1)

    # Assert
    backend._perform_memory_operations.assert_has_awaits([
        call([MemoryOperation(has_message_address, read_byte_count=1)]),
    ])
    assert len(backend.message_queue) > 0


@pytest.mark.asyncio
async def test_send_message_on_cooldown(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations.return_value = [b"\x00"]
    backend.message_cooldown = 2
    has_message_address = backend.patches.string_display.cstate_manager_global + 0x2

    # Run
    backend.display_message("Magoo")
    await backend._send_message_from_queue(1)

    # Assert
    backend._perform_memory_operations.assert_has_awaits([
        call([MemoryOperation(has_message_address, read_byte_count=1)]),
    ])
    assert len(backend.message_queue) > 0
    assert backend.message_cooldown == 1


@pytest.mark.asyncio
async def test_get_inventory(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations.return_value = [
        struct.pack(">II", item.index, item.index)
        for item in backend.game.resource_database.item
    ]

    # Run
    inventory = await backend._get_inventory()

    # Assert
    assert inventory == {
        item: InventoryItem(item.index, item.index)
        for item in backend.game.resource_database.item
    }


@pytest.mark.asyncio
async def test_check_for_collected_index_nothing(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._inventory = {
        backend.game.resource_database.multiworld_magic_item: InventoryItem(0, 0)
    }
    backend._update_inventory = AsyncMock()

    # Run
    await backend._check_for_collected_index()

    # Assert
    backend._update_inventory.assert_not_awaited()


@pytest.mark.asyncio
async def test_check_for_collected_index_location_collected_tracking(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend.tracking_inventory = True
    backend._inventory = {
        backend.game.resource_database.multiworld_magic_item: InventoryItem(10, 10)
    }
    backend._read_item = AsyncMock()
    backend._write_item = AsyncMock()
    backend._update_inventory = AsyncMock()
    backend._emit_location_collected = AsyncMock()
    # Run
    await backend._check_for_collected_index()

    # Assert
    backend._emit_location_collected.assert_awaited_once_with(9)
    backend._read_item.assert_not_awaited()
    backend._write_item.assert_not_awaited()
    backend._update_inventory.assert_awaited_once_with({
        backend.game.resource_database.multiworld_magic_item: InventoryItem(0, 0)
    })


@pytest.mark.asyncio
async def test_check_for_collected_index_location_collected_no_tracking(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend.tracking_inventory = False
    backend._inventory = None
    backend._read_item = AsyncMock(return_value=InventoryItem(10, 10))
    backend._write_item = AsyncMock()
    backend._update_inventory = AsyncMock()
    backend._emit_location_collected = AsyncMock()

    # Run
    await backend._check_for_collected_index()

    # Assert
    backend._emit_location_collected.assert_awaited_once_with(9)
    backend._update_inventory.assert_not_awaited()
    backend._read_item.assert_awaited_once_with(backend.game.resource_database.multiworld_magic_item)
    backend._write_item.assert_awaited_once_with(backend.game.resource_database.multiworld_magic_item,
                                                 InventoryItem(0, 0))


@pytest.mark.parametrize("tracking", [False, True])
@pytest.mark.asyncio
async def test_check_for_collected_index_receive_items(backend, tracking: bool):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend.tracking_inventory = tracking
    backend._update_inventory = AsyncMock()
    backend._read_item = AsyncMock(return_value=InventoryItem(0, 0))
    backend._write_item = AsyncMock()

    resource = backend.game.resource_database.energy_tank
    pickup = PickupEntry("Pickup", 0, ItemCategory.MISSILE, ItemCategory.MISSILE, (
        ConditionalResources(None, None,
                             ((resource, resource.max_capacity),),
                             ),
    ))
    inventory = {
        backend.game.resource_database.multiworld_magic_item: InventoryItem(0, 0),
        backend.game.resource_database.energy_tank: InventoryItem(1, 1),
    }
    backend._get_inventory = AsyncMock(return_value=inventory)
    if tracking:
        backend._inventory = inventory
    else:
        backend._inventory = None
    backend._permanent_pickups = [pickup]

    # Run
    await backend._check_for_collected_index()

    # Assert
    if tracking:
        backend._read_item.assert_not_awaited()
    else:
        backend._read_item.assert_awaited_once_with(backend.game.resource_database.multiworld_magic_item)
    backend._update_inventory.assert_awaited_once_with({
        backend.game.resource_database.multiworld_magic_item: InventoryItem(0, 1),
        backend.game.resource_database.energy_tank: InventoryItem(resource.max_capacity, resource.max_capacity),
    })
    backend._write_item.assert_not_awaited()


@pytest.mark.asyncio
async def test_update_inventory_no_change(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations = AsyncMock()
    backend._inventory = {
        item: InventoryItem(0, 0)
        for item in backend.game.resource_database.item
    }

    # Run
    await backend._update_inventory(backend._inventory)

    # Assert
    backend._perform_memory_operations.assert_not_awaited()


@pytest.mark.parametrize("item", [13])
@pytest.mark.asyncio
async def test_update_inventory_with_change(backend, item):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations = AsyncMock()
    backend._inventory = {
        item: InventoryItem(0, 0)
        for item in backend.game.resource_database.item
    }
    new_inventory = copy.copy(backend._inventory)
    new_inventory[backend.game.resource_database.multiworld_magic_item] = InventoryItem(1, 15)

    # Run
    await backend._update_inventory(new_inventory)

    # Assert
    backend._perform_memory_operations.assert_awaited_once_with([
        MemoryOperation(
            address=backend._get_player_state_pointer(),
            write_bytes=struct.pack(">II", 1, 15),
            read_byte_count=8,
            offset=connection_backend._powerup_offset(backend.game.resource_database.multiworld_magic_item.index),
        )
    ])


@pytest.mark.parametrize("query_result", [None, b"\x00" * 4])
@pytest.mark.asyncio
async def test_update_current_world_invalid(backend, query_result):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations = AsyncMock(return_value=[query_result])

    # Run
    await backend._update_current_world()

    # Assert
    assert backend._world is None


@pytest.mark.asyncio
async def test_update_current_world_present(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    world = backend.game.world_list.worlds[0]
    backend._perform_memory_operations = AsyncMock(return_value=[world.world_asset_id.to_bytes(4, "big")])

    # Run
    await backend._update_current_world()

    # Assert
    assert backend._world is world


@pytest.mark.parametrize("has_light_suit", [False, True])
@pytest.mark.asyncio
async def test_perform_write_inventory_dark_suit(backend, echoes_game_description, has_light_suit):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations = AsyncMock()
    dark_suit = echoes_game_description.resource_database.get_item(13)
    light_suit = echoes_game_description.resource_database.get_item(14)
    changed_items = {dark_suit: InventoryItem(1, 1)}
    if has_light_suit:
        backend._inventory = {light_suit: InventoryItem(1, 1)}
    else:
        backend._inventory = {}

    # Run
    await backend._perform_write_inventory(changed_items)

    # Assert
    backend._perform_memory_operations.assert_awaited_once()
    write_op = backend._perform_memory_operations.mock_calls[0].args[0]
    assert write_op[1] == MemoryOperation(
        address=2151533548,
        offset=84,
        write_bytes=b"\x00\x00\x00\x02" if has_light_suit else b"\x00\x00\x00\x01",
    )



import collections
from functools import partial
from typing import List, Dict, Optional

from PySide2 import QtCore, QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QRadioButton, QGroupBox, QHBoxLayout, QLabel, QPushButton, QVBoxLayout, \
    QApplication, QDialog, QAction, QMenu
from asyncqt import asyncSlot

from randovania.game_description import data_reader, default_database
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import PickupNode
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame
from randovania.games.prime import patcher_file
from randovania.gui.dialog.echoes_user_preferences_dialog import EchoesUserPreferencesDialog
from randovania.gui.dialog.game_input_dialog import GameInputDialog
from randovania.gui.generated.seed_details_window_ui import Ui_SeedDetailsWindow
from randovania.gui.lib import preset_describer, async_dialog, common_qt_lib
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin
from randovania.gui.lib.close_event_widget import CloseEventWidget
from randovania.gui.lib.common_qt_lib import set_default_window_icon, prompt_user_for_output_game_log
from randovania.gui.lib.window_manager import WindowManager
from randovania.interface_common import simplified_patcher, status_update_lib
from randovania.interface_common.options import Options, InfoAlert
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.interface_common.status_update_lib import ProgressUpdateCallable
from randovania.layout.layout_description import LayoutDescription


def _unique(iterable):
    seen = set()

    for item in iterable:
        if item in seen:
            continue
        seen.add(item)
        yield item


def _show_pickup_spoiler(button):
    target_player = getattr(button, "target_player", None)
    if target_player is not None:
        label = f"{button.item_name} for {button.player_names[target_player]}"
    else:
        label = button.item_name
    button.setText(label)
    button.item_is_hidden = False


def _hide_pickup_spoiler(button):
    button.setText("Hidden")
    button.item_is_hidden = True


class SeedDetailsWindow(CloseEventWidget, Ui_SeedDetailsWindow, BackgroundTaskMixin):
    _on_bulk_change: bool = False
    _history_items: List[QRadioButton]
    pickup_spoiler_buttons: List[QPushButton]
    layout_description: LayoutDescription
    _options: Options
    _window_manager: Optional[WindowManager]
    _player_names: Dict[int, str]
    _pickup_spoiler_current_game: Optional[GameDescription] = None

    def __init__(self, window_manager: Optional[WindowManager], options: Options):
        super().__init__()
        self.setupUi(self)
        set_default_window_icon(self)

        self._history_items = []
        self.pickup_spoiler_buttons = []
        self._pickup_spoiler_world_to_group = {}
        self._options = options
        self._window_manager = window_manager

        # Ui
        self._tool_button_menu = QMenu(self.tool_button)
        self.tool_button.setMenu(self._tool_button_menu)

        self._action_open_tracker = QAction(self)
        self._action_open_tracker.setText("Open map tracker")
        self._action_open_tracker.setEnabled(self._window_manager is not None)
        self._tool_button_menu.addAction(self._action_open_tracker)

        self._action_copy_permalink = QAction(self)
        self._action_copy_permalink.setText("Copy Permalink")
        self._tool_button_menu.addAction(self._action_copy_permalink)

        self._action_open_dolphin = QAction(self)
        self._action_open_dolphin.setText("Open Dolphin Hook")
        self._tool_button_menu.addAction(self._action_open_dolphin)
        self._action_open_dolphin.setVisible(False)

        # Signals
        self.export_log_button.clicked.connect(self._export_log)
        self.export_iso_button.clicked.connect(self._export_iso)
        self._action_open_tracker.triggered.connect(self._open_map_tracker)
        self._action_copy_permalink.triggered.connect(self._copy_permalink)
        self.pickup_spoiler_pickup_combobox.currentTextChanged.connect(self._on_change_pickup_filter)
        self.pickup_spoiler_show_all_button.clicked.connect(self._toggle_show_all_pickup_spoiler)
        self.player_index_combo.activated.connect(self._update_current_player)
        self.CloseEvent.connect(self.stop_background_process)

        # Progress
        self.background_tasks_button_lock_signal.connect(self.enable_buttons_with_background_tasks)
        self.progress_update_signal.connect(self.update_progress)
        self.stop_background_process_button.clicked.connect(self.stop_background_process)

        # Cosmetic
        self.customize_user_preferences_button.clicked.connect(self._open_user_preferences_dialog)

    @property
    def current_player_index(self) -> int:
        return self.player_index_combo.currentData()

    # Operations
    def _copy_permalink(self):
        QApplication.clipboard().setText(self.layout_description.permalink.as_base64_str)

    def _export_log(self):
        default_name = "Echoes Randomizer - {}.{}".format(self.layout_description.shareable_word_hash,
                                                          self.layout_description.file_extension())
        json_path = prompt_user_for_output_game_log(self, default_name=default_name)
        if json_path is not None:
            self.layout_description.save_to_file(json_path)

    async def _show_dialog_for_prime3_layout(self):
        from randovania.games.prime import gollop_corruption_patcher

        patches = self.layout_description.all_patches[self.current_player_index]
        game = default_database.game_description_for(RandovaniaGame.PRIME3)

        item_names = []
        for index in range(game.world_list.num_pickup_nodes):
            p_index = PickupIndex(index)
            if p_index in patches.pickup_assignment:
                name = patches.pickup_assignment[p_index].pickup.name
            else:
                name = "Missile Expansion"
            item_names.append(name)

        layout_string = gollop_corruption_patcher.layout_string_for_items(item_names)
        starting_location = patches.starting_location

        commands = "\n".join([
            f'set seed="{layout_string}"',
            f'set "starting_items={gollop_corruption_patcher.starting_items_for(patches.starting_items)}"',
            f'set "starting_location={gollop_corruption_patcher.starting_location_for(starting_location)}"',
        ])
        message_box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Information, "Commands for patcher",
                                            commands)
        common_qt_lib.set_default_window_icon(message_box)
        message_box.setTextInteractionFlags(Qt.TextSelectableByMouse)
        QApplication.clipboard().setText(commands)
        await async_dialog.execute_dialog(message_box)

    @asyncSlot()
    async def _export_iso(self):
        layout = self.layout_description
        has_spoiler = layout.permalink.spoiler
        options = self._options

        if layout.permalink.get_preset(self.current_player_index).configuration.game == RandovaniaGame.PRIME3:
            return await self._show_dialog_for_prime3_layout()

        if not options.is_alert_displayed(InfoAlert.FAQ):
            await async_dialog.message_box(self, QtWidgets.QMessageBox.Icon.Information, "FAQ",
                                           "Have you read the Randovania FAQ?\n"
                                           "It can be found in the main Randovania window → Help → FAQ")
            options.mark_alert_as_displayed(InfoAlert.FAQ)

        dialog = GameInputDialog(options, "Echoes Randomizer - {}.iso".format(layout.shareable_word_hash), has_spoiler)
        result = await async_dialog.execute_dialog(dialog)

        if result != QDialog.Accepted:
            return

        if simplified_patcher.export_busy:
            return await async_dialog.message_box(
                self, QtWidgets.QMessageBox.Critical,
                "Can't save ISO",
                "Error: Unable to save multiple ISOs at the same time,"
                "another window is saving an ISO right now.")

        input_file = dialog.input_file
        output_file = dialog.output_file
        auto_save_spoiler = dialog.auto_save_spoiler
        player_index = self.current_player_index
        player_names = self._player_names

        with options:
            options.output_directory = output_file.parent
            options.auto_save_spoiler = auto_save_spoiler

        def work(progress_update: ProgressUpdateCallable):
            num_updaters = 2
            if input_file is not None:
                num_updaters += 1
            updaters = status_update_lib.split_progress_update(progress_update, num_updaters)

            if input_file is not None:
                simplified_patcher.unpack_iso(input_iso=input_file,
                                              options=options,
                                              progress_update=updaters[0])

            # Apply Layout
            simplified_patcher.apply_layout(layout=layout,
                                            players_config=PlayersConfiguration(
                                                player_index=player_index,
                                                player_names=player_names,
                                            ),
                                            options=options,
                                            progress_update=updaters[-2])

            # Pack ISO
            simplified_patcher.pack_iso(output_iso=output_file,
                                        options=options,
                                        progress_update=updaters[-1])
            if has_spoiler and auto_save_spoiler:
                layout.save_to_file(output_file.with_suffix(f".{LayoutDescription.file_extension()}"))

            progress_update(f"Finished!", 1)

        self.run_in_background_thread(work, "Exporting...")

    def _open_map_tracker(self):
        current_preset = self.layout_description.permalink.presets[self.current_player_index]
        self._window_manager.open_map_tracker(current_preset.configuration)

    # Layout Visualization
    def _create_pickup_spoilers(self, game_description: GameDescription):
        if game_description == self._pickup_spoiler_current_game:
            return

        for groups in self._pickup_spoiler_world_to_group.values():
            groups.deleteLater()

        self._pickup_spoiler_current_game = game_description
        self.pickup_spoiler_show_all_button.currently_show_all = True
        self.pickup_spoiler_buttons.clear()

        self._pickup_spoiler_world_to_group = {}
        nodes_in_world = collections.defaultdict(list)

        for world, area, node in game_description.world_list.all_worlds_areas_nodes:
            if isinstance(node, PickupNode):
                world_name = world.correct_name(area.in_dark_aether)
                nodes_in_world[world_name].append((f"{area.name} - {node.name}", node.pickup_index))
                continue

        for world_name in sorted(nodes_in_world.keys()):
            group_box = QGroupBox(self.pickup_spoiler_scroll_contents)
            group_box.setTitle(world_name)
            vertical_layout = QVBoxLayout(group_box)
            vertical_layout.setContentsMargins(8, 4, 8, 4)
            vertical_layout.setSpacing(2)
            group_box.vertical_layout = vertical_layout

            vertical_layout.horizontal_layouts = []
            self._pickup_spoiler_world_to_group[world_name] = group_box
            self.pickup_spoiler_scroll_content_layout.addWidget(group_box)

            for area_name, pickup_index in sorted(nodes_in_world[world_name], key=lambda it: it[0]):
                horizontal_layout = QHBoxLayout()
                horizontal_layout.setSpacing(2)

                label = QLabel(group_box)
                label.setText(area_name)
                horizontal_layout.addWidget(label)
                horizontal_layout.label = label

                push_button = QPushButton(group_box)
                push_button.setFlat(True)
                push_button.setText("Hidden")
                push_button.item_is_hidden = True
                push_button.pickup_index = pickup_index
                push_button.clicked.connect(partial(self._toggle_pickup_spoiler, push_button))
                push_button.item_name = "Nothing was Set, ohno"
                push_button.row = horizontal_layout
                horizontal_layout.addWidget(push_button)
                horizontal_layout.button = push_button
                self.pickup_spoiler_buttons.append(push_button)

                group_box.vertical_layout.addLayout(horizontal_layout)
                group_box.vertical_layout.horizontal_layouts.append(horizontal_layout)

    def create_history_item(self, node):
        button = QRadioButton(self.layout_history_content)
        button.toggled.connect(self.on_select_node)
        button.setText(node)
        self.layout_history_content_layout.addWidget(button)
        self._history_items.append(button)
        return button

    def update_layout_description(self, description: LayoutDescription):
        self.layout_description = description
        self.layout_info_tab.show()

        self.setWindowTitle(f"Game Details: {description.shareable_word_hash}")
        self.export_log_button.setEnabled(description.permalink.spoiler)

        self._player_names = {
            i: f"Player {i + 1}"
            for i in range(description.permalink.player_count)
        }

        self.export_iso_button.setEnabled(description.permalink.player_count == 1)
        if description.permalink.player_count > 1:
            self.export_iso_button.setToolTip("Multiworld games can only be exported from a game session")
        else:
            self.export_iso_button.setToolTip("")

        self.customize_user_preferences_button.setVisible(description.permalink.player_count == 1)

        self.player_index_combo.clear()
        for i in range(description.permalink.player_count):
            self.player_index_combo.addItem(self._player_names[i], i)
        self.player_index_combo.setCurrentIndex(0)
        self.player_index_combo.setVisible(description.permalink.player_count > 1)

        self._update_current_player()

    def _update_current_player(self):
        description = self.layout_description
        current_player = self.current_player_index
        preset = description.permalink.get_preset(current_player)

        self.permalink_edit.setText(description.permalink.as_base64_str)
        title_text = """
        <p>
            Seed Hash: {description.shareable_word_hash} ({description.shareable_hash})<br/>
            Preset Name: {preset.name}
        </p>
        """.format(description=description, preset=preset)
        self.layout_title_label.setText(title_text)

        categories = list(preset_describer.describe(preset))
        self.layout_description_left_label.setText(preset_describer.merge_categories(categories[::2]))
        self.layout_description_right_label.setText(preset_describer.merge_categories(categories[1::2]))

        # Game Spoiler
        has_spoiler = description.permalink.spoiler
        self.pickup_tab.setEnabled(has_spoiler)
        patches = description.all_patches[current_player]

        if has_spoiler:
            pickup_names = {
                pickup.pickup.name
                for pickup in patches.pickup_assignment.values()
            }
            game_description = data_reader.decode_data(preset.configuration.game_data)
            self._create_pickup_spoilers(game_description)
            starting_area = game_description.world_list.area_by_area_location(patches.starting_location)

            extra_items = patcher_file.additional_starting_items(preset.configuration,
                                                                 game_description.resource_database,
                                                                 patches.starting_items)

            self.spoiler_starting_location_label.setText("Starting Location: {}".format(
                game_description.world_list.area_name(starting_area)
            ))
            self.spoiler_starting_items_label.setText("Random Starting Items: {}".format(
                ", ".join(extra_items)
                if extra_items else "None"
            ))

        else:
            pickup_names = {}
            self.layout_info_tab.removeTab(self.layout_info_tab.indexOf(self.pickup_tab))
            self.spoiler_starting_location_label.setText("Starting Location")
            self.spoiler_starting_items_label.setText("Random Starting Items")

        self.pickup_spoiler_pickup_combobox.clear()
        self.pickup_spoiler_pickup_combobox.addItem("None")
        for pickup_name in sorted(pickup_names):
            self.pickup_spoiler_pickup_combobox.addItem(pickup_name)

        for pickup_button in self.pickup_spoiler_buttons:
            pickup_target = patches.pickup_assignment.get(pickup_button.pickup_index)

            pickup_button.target_player = None
            if pickup_target is not None:
                pickup_button.item_name = pickup_target.pickup.name if has_spoiler else "????"
                if has_spoiler and description.permalink.player_count > 1:
                    pickup_button.target_player = pickup_target.player
                    pickup_button.player_names = self._player_names
            else:
                pickup_button.item_name = "Nothing"

            if not pickup_button.item_is_hidden:
                pickup_button.setText(pickup_button.item_name)

    def _toggle_pickup_spoiler(self, button):
        if button.item_is_hidden:
            _show_pickup_spoiler(button)
        else:
            _hide_pickup_spoiler(button)
        self._update_show_all_button_state()

    def _toggle_show_all_pickup_spoiler(self):
        if self.pickup_spoiler_show_all_button.currently_show_all:
            action = _show_pickup_spoiler
        else:
            action = _hide_pickup_spoiler

        for button in self.pickup_spoiler_buttons:
            action(button)

        self._update_show_all_button_state()

    def _update_show_all_button_state(self):
        self.pickup_spoiler_show_all_button.currently_show_all = all(
            button.item_is_hidden for button in self.pickup_spoiler_buttons
        )
        if self.pickup_spoiler_show_all_button.currently_show_all:
            self.pickup_spoiler_show_all_button.setText(QtCore.QCoreApplication.translate("HistoryWindow", "Show All"))
        else:
            self.pickup_spoiler_show_all_button.setText(QtCore.QCoreApplication.translate("HistoryWindow", "Hide All"))

    def _on_change_pickup_filter(self, text):
        for button in self.pickup_spoiler_buttons:
            visible = text == "None" or text == button.item_name
            button.setVisible(visible)
            button.row.label.setVisible(visible)

    def _open_user_preferences_dialog(self):
        dialog = EchoesUserPreferencesDialog(self, self._options.cosmetic_patches)
        result = dialog.exec_()
        if result == QDialog.Accepted:
            with self._options as options:
                options.cosmetic_patches = dialog.cosmetic_patches

    def enable_buttons_with_background_tasks(self, value: bool):
        self.stop_background_process_button.setEnabled(not value)
        self.export_iso_button.setEnabled(value)
        simplified_patcher.export_busy = not value

    def update_progress(self, message: str, percentage: int):
        self.progress_label.setText(message)
        if "Aborted" in message:
            percentage = 0
        if percentage >= 0:
            self.progress_bar.setRange(0, 100)
            self.progress_bar.setValue(percentage)
        else:
            self.progress_bar.setRange(0, 0)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_item_pool.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_PresetItemPool(object):
    def setupUi(self, PresetItemPool):
        if not PresetItemPool.objectName():
            PresetItemPool.setObjectName(u"PresetItemPool")
        PresetItemPool.resize(566, 450)
        PresetItemPool.setMaximumSize(QSize(16777215, 16777215))
        self.centralWidget = QWidget(PresetItemPool)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.scroll_area = QScrollArea(self.centralWidget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setFrameShape(QFrame.NoFrame)
        self.scroll_area.setFrameShadow(QFrame.Plain)
        self.scroll_area.setLineWidth(0)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 566, 450))
        self.gridLayout = QGridLayout(self.scroll_area_contents)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.item_alternative_box = QGroupBox(self.scroll_area_contents)
        self.item_alternative_box.setObjectName(u"item_alternative_box")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.item_alternative_box.sizePolicy().hasHeightForWidth())
        self.item_alternative_box.setSizePolicy(sizePolicy)
        self.item_alternative_layout = QVBoxLayout(self.item_alternative_box)
        self.item_alternative_layout.setSpacing(6)
        self.item_alternative_layout.setContentsMargins(11, 11, 11, 11)
        self.item_alternative_layout.setObjectName(u"item_alternative_layout")

        self.gridLayout.addWidget(self.item_alternative_box, 0, 0, 1, 2)

        self.item_pool_box = QGroupBox(self.scroll_area_contents)
        self.item_pool_box.setObjectName(u"item_pool_box")
        self.item_pool_box.setToolTipDuration(-1)
        self.item_pool_box.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.item_pool_box.setFlat(False)
        self.gridLayout_3 = QGridLayout(self.item_pool_box)
        self.gridLayout_3.setSpacing(6)
        self.gridLayout_3.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.ammo_box = QGroupBox(self.item_pool_box)
        self.ammo_box.setObjectName(u"ammo_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.ammo_box.sizePolicy().hasHeightForWidth())
        self.ammo_box.setSizePolicy(sizePolicy1)
        self.ammo_box.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.ammo_layout = QVBoxLayout(self.ammo_box)
        self.ammo_layout.setSpacing(6)
        self.ammo_layout.setContentsMargins(11, 11, 11, 11)
        self.ammo_layout.setObjectName(u"ammo_layout")
        self.ammo_layout.setSizeConstraint(QLayout.SetMinimumSize)

        self.gridLayout_3.addWidget(self.ammo_box, 1, 1, 1, 1)

        self.major_items_box = QGroupBox(self.item_pool_box)
        self.major_items_box.setObjectName(u"major_items_box")
        sizePolicy1.setHeightForWidth(self.major_items_box.sizePolicy().hasHeightForWidth())
        self.major_items_box.setSizePolicy(sizePolicy1)
        self.major_items_box.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.major_items_layout = QGridLayout(self.major_items_box)
        self.major_items_layout.setSpacing(6)
        self.major_items_layout.setContentsMargins(11, 11, 11, 11)
        self.major_items_layout.setObjectName(u"major_items_layout")
        self.major_items_layout.setSizeConstraint(QLayout.SetMinimumSize)

        self.gridLayout_3.addWidget(self.major_items_box, 1, 0, 1, 1)

        self.item_pool_count_label = QLabel(self.item_pool_box)
        self.item_pool_count_label.setObjectName(u"item_pool_count_label")
        self.item_pool_count_label.setAlignment(Qt.AlignCenter)

        self.gridLayout_3.addWidget(self.item_pool_count_label, 0, 0, 1, 2)


        self.gridLayout.addWidget(self.item_pool_box, 3, 0, 1, 2)

        self.random_starting_box = QGroupBox(self.scroll_area_contents)
        self.random_starting_box.setObjectName(u"random_starting_box")
        self.gridLayout_2 = QGridLayout(self.random_starting_box)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.maximum_starting_label = QLabel(self.random_starting_box)
        self.maximum_starting_label.setObjectName(u"maximum_starting_label")

        self.gridLayout_2.addWidget(self.maximum_starting_label, 2, 0, 1, 1)

        self.minimum_starting_label = QLabel(self.random_starting_box)
        self.minimum_starting_label.setObjectName(u"minimum_starting_label")
        self.minimum_starting_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)

        self.gridLayout_2.addWidget(self.minimum_starting_label, 1, 0, 1, 1)

        self.minimum_starting_spinbox = QSpinBox(self.random_starting_box)
        self.minimum_starting_spinbox.setObjectName(u"minimum_starting_spinbox")
        self.minimum_starting_spinbox.setMaximum(30)

        self.gridLayout_2.addWidget(self.minimum_starting_spinbox, 1, 1, 1, 1)

        self.random_starting_label = QLabel(self.random_starting_box)
        self.random_starting_label.setObjectName(u"random_starting_label")
        self.random_starting_label.setWordWrap(True)

        self.gridLayout_2.addWidget(self.random_starting_label, 0, 0, 1, 2)

        self.maximum_starting_spinbox = QSpinBox(self.random_starting_box)
        self.maximum_starting_spinbox.setObjectName(u"maximum_starting_spinbox")
        self.maximum_starting_spinbox.setMaximum(30)

        self.gridLayout_2.addWidget(self.maximum_starting_spinbox, 2, 1, 1, 1)


        self.gridLayout.addWidget(self.random_starting_box, 2, 0, 1, 2)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.verticalLayout.addWidget(self.scroll_area)

        PresetItemPool.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetItemPool)

        QMetaObject.connectSlotsByName(PresetItemPool)
    # setupUi

    def retranslateUi(self, PresetItemPool):
        PresetItemPool.setWindowTitle(QCoreApplication.translate("PresetItemPool", u"Item Pool", None))
        self.item_alternative_box.setTitle(QCoreApplication.translate("PresetItemPool", u"Item Alternatives", None))
        self.item_pool_box.setTitle(QCoreApplication.translate("PresetItemPool", u"Item Pool", None))
        self.ammo_box.setTitle(QCoreApplication.translate("PresetItemPool", u"Ammo", None))
        self.major_items_box.setTitle(QCoreApplication.translate("PresetItemPool", u"Major Items", None))
#if QT_CONFIG(tooltip)
        self.item_pool_count_label.setToolTip(QCoreApplication.translate("PresetItemPool", u"If there are fewer than 119 items, the rest of the item locations will contain Energy Transfer Modules.", None))
#endif // QT_CONFIG(tooltip)
        self.item_pool_count_label.setText(QCoreApplication.translate("PresetItemPool", u"Items in pool: #/119", None))
        self.random_starting_box.setTitle(QCoreApplication.translate("PresetItemPool", u"Random Starting Items", None))
        self.maximum_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"Start with at most this many items:", None))
        self.minimum_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"Start with at least this many items:", None))
        self.random_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"<html><head/><body><p>Randovania will add additional starting items if necessary to make the seed possible.<br/>The first value controls how many items are always added.<br/>The second value controls how many items the seed can have before it fails to generate.</p></body></html>", None))
    # retranslateUi




from pathlib import Path
from typing import Optional, Dict

import PySide2
import slugify
from PySide2.QtCore import QTimer, Signal, Qt
from PySide2.QtGui import QPixmap
from PySide2.QtWidgets import QMainWindow, QLabel
from asyncqt import asyncSlot

from randovania import get_data_path
from randovania.game_connection.connection_base import GameConnectionStatus, InventoryItem
from randovania.game_connection.game_connection import GameConnection
from randovania.game_description import data_reader
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_database import find_resource_info_with_long_name
from randovania.games.prime import default_data
from randovania.gui.generated.auto_tracker_window_ui import Ui_AutoTrackerWindow
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.clickable_label import ClickableLabel
from randovania.gui.lib.game_connection_setup import GameConnectionSetup
from randovania.gui.lib.pixmap_lib import paint_with_opacity
from randovania.interface_common.options import Options


class AutoTrackerWindow(QMainWindow, Ui_AutoTrackerWindow):
    _hooked = False
    _base_address: int
    give_item_signal = Signal(PickupEntry)

    def __init__(self, game_connection: GameConnection, options: Options):
        super().__init__()
        self.setupUi(self)
        self.game_connection = game_connection
        common_qt_lib.set_default_window_icon(self)

        self.game_data = data_reader.decode_data(default_data.decode_default_prime2())
        self._energy_tank_item = find_resource_info_with_long_name(self.game_data.resource_database.item, "Energy Tank")

        self._item_to_label: Dict[ItemResourceInfo, ClickableLabel] = {}
        self._labels_for_keys = []
        self.create_tracker()

        self.game_connection_setup = GameConnectionSetup(self, self.game_connection_tool, self.connection_status_label,
                                                         self.game_connection, options)
        self.force_update_button.setEnabled(not options.tracking_inventory)
        self.force_update_button.clicked.connect(self.on_force_update_button)

        self._update_timer = QTimer(self)
        self._update_timer.setInterval(100)
        self._update_timer.timeout.connect(self._on_timer_update)
        self._update_timer.setSingleShot(True)

    def showEvent(self, event: PySide2.QtGui.QShowEvent):
        self._update_timer.start()
        super().showEvent(event)

    def hideEvent(self, event: PySide2.QtGui.QHideEvent):
        self._update_timer.stop()
        super().hideEvent(event)

    def _update_tracker_from_hook(self, inventory: Dict[ItemResourceInfo, InventoryItem]):
        for item, label in self._item_to_label.items():
            current = inventory.get(item, InventoryItem(0, 0))
            label.set_checked(current.capacity > 0)

        energy_tank = inventory.get(self._energy_tank_item, InventoryItem(0, 0))
        self._energy_tank_label.setText("x {}/{}".format(energy_tank.capacity, self._energy_tank_item.max_capacity))

        for label, keys in self._labels_for_keys:
            num_keys = sum(inventory.get(key, InventoryItem(0, 0)).capacity for key in keys)
            label.setText("x {}/{}".format(num_keys, len(keys)))

    @asyncSlot()
    async def _on_timer_update(self):
        try:
            self.force_update_button.setEnabled(not self.game_connection.tracking_inventory)
            current_status = self.game_connection.current_status
            if current_status == GameConnectionStatus.InGame or current_status == GameConnectionStatus.TrackerOnly:
                inventory = self.game_connection.get_current_inventory()
                self._update_tracker_from_hook(inventory)
        finally:
            self._update_timer.start()

    def create_tracker(self):
        def get_image_path(image_name: str) -> Path:
            return get_data_path().joinpath(f"gui_assets/tracker/images-noanim/{image_name}.gif")

        def find_resource(name: str):
            return find_resource_info_with_long_name(self.game_data.resource_database.item, name)

        def create_item(image_name: str, always_opaque: bool = False):
            image_path = get_image_path(image_name)
            pixmap = QPixmap(str(image_path))

            label = ClickableLabel(self.inventory_group, paint_with_opacity(pixmap, 0.3),
                                   paint_with_opacity(pixmap, 1.0))
            label.set_checked(always_opaque)
            label.set_ignore_mouse_events(always_opaque)

            return label

        def add_item(row: int, column: int, item_name: str, image_name: Optional[str] = None):
            if image_name is None:
                image_name = slugify.slugify(item_name).replace("-", "_")

            label = create_item(image_name)
            label.set_ignore_mouse_events(True)
            self.inventory_layout.addWidget(label, row, column)

            self._item_to_label[find_resource(item_name)] = label

        add_item(0, 0, "Charge Beam")
        add_item(1, 0, "Missile Launcher")
        add_item(2, 0, "Super Missile")
        add_item(3, 0, "Seeker Launcher")

        add_item(0, 1, "Dark Beam")
        add_item(0, 2, "Light Beam")
        add_item(0, 3, "Annihilator Beam")
        add_item(1, 1, "Darkburst")
        add_item(1, 2, "Sunburst")
        add_item(1, 3, "Sonic Boom")

        add_item(2, 2, "Scan Visor")
        add_item(2, 3, "Dark Visor")
        add_item(2, 4, "Echo Visor")

        add_item(3, 1, "Space Jump Boots")
        add_item(3, 2, "Gravity Boost")
        add_item(3, 3, "Grapple Beam")
        add_item(3, 4, "Screw Attack")

        add_item(0, 4, "Dark Suit")
        add_item(1, 4, "Light Suit")

        add_item(0, 5, "Morph Ball Bomb")
        add_item(1, 5, "Power Bomb")
        add_item(2, 5, "Boost Ball")
        add_item(3, 5, "Spider Ball")

        add_item(0, 6, "Violet Translator")
        add_item(1, 6, "Amber Translator")
        add_item(2, 6, "Emerald Translator")
        add_item(3, 6, "Cobalt Translator")

        add_item(2, 1, "Energy Transfer Module")

        self.inventory_layout.addWidget(create_item("energy_tank", True), 5, 1)
        energy_tank_label = QLabel(self.inventory_group)
        energy_tank_label.setText("x 0/14")
        energy_tank_label.setAlignment(Qt.AlignCenter)
        self.inventory_layout.addWidget(energy_tank_label, 6, 1)
        self._energy_tank_label = energy_tank_label

        self.inventory_layout.addWidget(create_item("dark_agon_key-recolored", True), 5, 2)
        dark_agon_key_label = QLabel(self.inventory_group)
        dark_agon_key_label.setText("x 0/3")
        dark_agon_key_label.setAlignment(Qt.AlignCenter)
        self.inventory_layout.addWidget(dark_agon_key_label, 6, 2)
        self._labels_for_keys.append((
            dark_agon_key_label,
            (find_resource("Dark Agon Key 1"), find_resource("Dark Agon Key 2"), find_resource("Dark Agon Key 3"))
        ))

        self.inventory_layout.addWidget(create_item("dark_torvus_key-recolored", True), 5, 3)
        dark_torvus_key_label = QLabel(self.inventory_group)
        dark_torvus_key_label.setText("x 0/3")
        dark_torvus_key_label.setAlignment(Qt.AlignCenter)
        self.inventory_layout.addWidget(dark_torvus_key_label, 6, 3)
        self._labels_for_keys.append((
            dark_torvus_key_label,
            (find_resource("Dark Torvus Key 1"), find_resource("Dark Torvus Key 2"), find_resource("Dark Torvus Key 3"))
        ))

        self.inventory_layout.addWidget(create_item("ing_hive_key-recolored", True), 5, 4)
        ing_hive_key_label = QLabel(self.inventory_group)
        ing_hive_key_label.setText("x 0/3")
        ing_hive_key_label.setAlignment(Qt.AlignCenter)
        self.inventory_layout.addWidget(ing_hive_key_label, 6, 4)
        self._labels_for_keys.append((
            ing_hive_key_label,
            (find_resource("Ing Hive Key 1"), find_resource("Ing Hive Key 2"), find_resource("Ing Hive Key 3"))
        ))

        self.inventory_layout.addWidget(create_item("sky_temple_key", True), 5, 5)
        sky_temple_key_label = QLabel(self.inventory_group)
        sky_temple_key_label.setText("x 0/9")
        sky_temple_key_label.setAlignment(Qt.AlignCenter)
        self.inventory_layout.addWidget(sky_temple_key_label, 6, 5)
        self._labels_for_keys.append((
            sky_temple_key_label,
            (find_resource("Sky Temple Key 1"), find_resource("Sky Temple Key 2"), find_resource("Sky Temple Key 3"),
             find_resource("Sky Temple Key 4"), find_resource("Sky Temple Key 5"), find_resource("Sky Temple Key 6"),
             find_resource("Sky Temple Key 7"), find_resource("Sky Temple Key 8"), find_resource("Sky Temple Key 9"),)
        ))

    @asyncSlot()
    async def on_force_update_button(self):
        await self.game_connection.backend.update_current_inventory()



import argparse
import collections
import copy
import csv
import json
import os
import re
from pathlib import Path
from statistics import stdev
from typing import Dict, Tuple, Optional

from randovania.game_description.default_database import game_description_for
from randovania.game_description.node import PickupNode, LogbookNode
from randovania.games.game import RandovaniaGame


def read_json(path: Path) -> dict:
    with path.open() as x:
        return json.load(x)


_KEY_MATCH = re.compile(r"Key (\d+)")
_PLAYER_MATCH = re.compile(r" for Player \d+")


def _filter_item_name(name: str) -> str:
    return _PLAYER_MATCH.sub("", _KEY_MATCH.sub("Key", name))


def accumulate_results(game_modifications: dict,
                       items: Dict[str, Dict[str, int]],
                       locations: Dict[str, Dict[str, int]],
                       item_hints: Dict[str, Dict[str, int]],
                       location_hints: Dict[str, Dict[str, int]],

                       index_to_location: Dict[int, Tuple[str, str]],
                       logbook_to_name: Dict[str, str],
                       ):
    for worlds, world_data in game_modifications["locations"].items():
        for area_name, item_name in world_data.items():
            item_name = _filter_item_name(item_name)
            items[item_name][area_name] += 1
            locations[area_name][item_name] += 1

    for logbook_asset, hint_data in game_modifications["hints"].items():
        if hint_data["hint_type"] != "location":
            continue

        logbook_asset = logbook_to_name[logbook_asset]

        if hint_data["target"] == -1:
            item_name = "Nothing"
        else:
            area_name, location_name = index_to_location[hint_data["target"]]
            item_name = game_modifications["locations"][area_name][location_name]

        item_name = _filter_item_name(item_name)
        location_hints[logbook_asset][item_name] += 1
        item_hints[item_name][logbook_asset] += 1


def calculate_pickup_count(items: Dict[str, Dict[str, int]]) -> Dict[str, int]:
    return {
        name: sum(data.values())
        for name, data in items.items()
    }


def sort_by_contents(data: dict) -> dict:
    return {
        item: {
            location: count
            for location, count in sorted(data[item].items(), key=lambda t: t[1], reverse=True)
        }
        for item in sorted(data.keys())
    }


def calculate_stddev(pickup_count: Dict[str, int], item_counts: Dict[str, float]) -> float:
    balanced_freq = {
        item: count / pickup_count[item]
        for item, count in item_counts.items()
    }
    return stdev(balanced_freq.values())


def first_key(d: dict):
    for key in d:
        return key


def create_report(seeds_dir: str, output_file: str, csv_dir: Optional[str]):
    def item_creator():
        return collections.defaultdict(int)

    items = collections.defaultdict(item_creator)
    locations = collections.defaultdict(item_creator)
    item_hints = collections.defaultdict(item_creator)
    location_hints = collections.defaultdict(item_creator)

    game_description = game_description_for(RandovaniaGame.PRIME2)
    world_list = game_description.world_list
    index_to_location = {
        node.pickup_index.index: (world_list.world_name_from_node(node, distinguish_dark_aether=True),
                                  world_list.node_name(node))
        for node in game_description.world_list.all_nodes
        if isinstance(node, PickupNode)
    }

    logbook_to_name = {
        str(node.string_asset_id): game_description.world_list.node_name(node)
        for node in game_description.world_list.all_nodes
        if isinstance(node, LogbookNode)
    }

    seed_count = 0
    pickup_count = None
    for seed in Path(seeds_dir).glob("**/*.json"):
        for game_modifications in read_json(seed)["game_modifications"]:
            accumulate_results(game_modifications,
                               items, locations,
                               item_hints, location_hints,
                               index_to_location, logbook_to_name)
        if seed_count == 0:
            pickup_count = calculate_pickup_count(items)
        seed_count += 1

    if pickup_count is None:
        raise Exception("No seeds found")

    stddev_by_location = {
        location: calculate_stddev(pickup_count, locations[location])
        for location in locations.keys()
    }

    final_results = {
        "seed_count": seed_count,
        "stddev_by_location": {
            location: stddev
            for location, stddev in sorted(stddev_by_location.items(), key=lambda t: t[1], reverse=True)
        },
        "items": sort_by_contents(items),
        "locations": sort_by_contents(locations),
        "item_hints": sort_by_contents(item_hints),
        "location_hints": sort_by_contents(location_hints),
    }

    if csv_dir is not None:
        os.makedirs(csv_dir, exist_ok=True)
        for field in "items", "locations", "item_hints", "location_hints":
            data = final_results[field]

            possible_columns = set()
            for potential_values in data.values():
                possible_columns |= set(potential_values.keys())

            possible_columns = list(sorted(possible_columns))
            possible_columns.insert(0, "row_name")

            with open(os.path.join(csv_dir, field + ".csv"), "w", newline='') as csv_file:
                writer = csv.DictWriter(csv_file, fieldnames=possible_columns)
                writer.writeheader()

                for column, row_data in data.items():
                    row_data = copy.copy(row_data)
                    row_data["row_name"] = column
                    writer.writerow(row_data)

    with open(output_file, "w") as output:
        json.dump(final_results, output, indent=4, separators=(',', ': '))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--csv-dir")
    parser.add_argument("seeds_dir")
    parser.add_argument("output_file")
    args = parser.parse_args()
    create_report(args.seeds_dir, args.output_file,
                  args.csv_dir)


if __name__ == "__main__":
    main()



import argparse
import asyncio
import os
import sys

from PySide2.QtCore import QCoreApplication

from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_connection.dolphin_backend import DolphinBackend
from randovania.game_connection.game_connection import GameConnection


should_quit = False
old_hook = sys.excepthook


async def worker(app: QCoreApplication):
    connection = GameConnection(DolphinBackend())

    await connection.start()
    try:
        while connection.current_status != GameConnectionStatus.InGame:
            await asyncio.sleep(1)
            print(connection.pretty_current_status)

        await connection.display_message("Hello Samus.")
        await connection.display_message("I see you are trying to do a multiworld.")
        await connection.display_message("But are you authorized for that?")
        await asyncio.sleep(20)

    except KeyboardInterrupt:
        return

    finally:
        await connection.stop()


def main():
    parser = argparse.ArgumentParser()
    parser.parse_args()

    app = QCoreApplication(sys.argv)

    os.environ['QT_API'] = "PySide2"
    import asyncqt
    loop = asyncqt.QEventLoop(app)
    asyncio.set_event_loop(loop)

    def catch_exceptions(t, val, tb):
        global should_quit
        should_quit = True
        old_hook(t, val, tb)

    sys.excepthook = catch_exceptions

    with loop:
        try:
            asyncio.get_event_loop().run_until_complete(worker(app))
        finally:
            app.quit()


if __name__ == '__main__':
    main()
